#include "libs.h"
#include "model_body.h"
#include "space.h"
#include "matrix4x4.h"
#include "frame.h"
#include "l3d.h"
#include "world_view.h"
#include "model_coll_mesh_data.h"

ModelBody::ModelBody(void): Body() {
  triMeshLastMatrixIndex = 0;
}

ModelBody::~ModelBody(void) {
  SetFrame(0); /* Will remove geom from frame if necessary. */
  for(unsigned int i = 0; i < geoms.size(); i++) {
    dGeomDestroy(geoms[i]);
  }
}

void ModelBody::Disable(void) {
  for(unsigned int i = 0; i < geoms.size(); i++) {
    dGeomDisable(geoms[i]);
  }
}

void ModelBody::Enable(void) {
  for(unsigned int i = 0; i < geoms.size(); i++) {
    dGeomEnable(geoms[i]);
  }
}

void ModelBody::GeomsSetBody(dBodyID body) {
  for(unsigned int i = 0; i < geoms.size(); i++) {
    dGeomSetBody(geoms[i], body);
  }
}

void ModelBody::SetModel(int sbreModel) {
  assert(geoms.size() == 0);
  CollMeshSet* mset = GetModelCollMeshSet(sbreModel);

  geomColl.resize(mset->numMeshParts);
  geoms.resize(mset->numMeshParts);

  for(int i = 0; i < mset->numMeshParts; i++) {
    geoms[i] = dCreateTriMesh(0, mset->meshParts[i], NULL, NULL, NULL);
    geomColl[i].parent = this;
    geomColl[i].flags = mset->meshInfo[i].flags;
    dGeomSetData(geoms[i], static_cast<Object*>(&geomColl[i]));
  }
}

void ModelBody::SetPosition(vector3d p) {
  for(unsigned int i = 0; i < geoms.size(); i++) {
    dGeomSetPosition(geoms[i], p.x, p.y, p.z);
  }
}

vector3d ModelBody::GetPosition(void) {
  const dReal* pos = dGeomGetPosition(geoms[0]);
  return vector3d(pos[0], pos[1], pos[2]);
}

void ModelBody::SetRotation(const matrix4x4d& r) {
  dMatrix3 _m;
  r.SaveToOdeMatrix(_m);
  for(unsigned int i = 0; i < geoms.size(); i++) {
    dGeomSetRotation(geoms[i], _m);
  }
}

void ModelBody::GetRotMatrix(matrix4x4d& m) {
  m.LoadFromOdeMatrix(dGeomGetRotation(geoms[0]));
}

void ModelBody::ViewingRotation(void) {
  matrix4x4d m;
  GetRotMatrix(m);
  m = m.InverseOf();
  glMultMatrixd(&m[0]);
}

void ModelBody::TransformToModelCoords(const Frame* camFrame) {
  vector3d fpos = GetPositionRelTo(camFrame);

  const dReal* r = dGeomGetRotation(geoms[0]);
  matrix4x4d m;
  m[ 0] = r[ 0]; m[ 1] = r[ 4]; m[ 2] = r[ 8]; m[ 3] = 0;
  m[ 4] = r[ 1]; m[ 5] = r[ 5]; m[ 6] = r[ 9]; m[ 7] = 0;
  m[ 8] = r[ 2]; m[ 9] = r[ 6]; m[10] = r[10]; m[11] = 0;
  m[12] = fpos.x; m[13] = fpos.y; m[14] = fpos.z; m[15] = 1;
  glMultMatrixd(&m[0]);
}

void ModelBody::SetFrame(Frame* f) {
  if(GetFrame()) {
    for(unsigned int i = 0; i < geoms.size(); i++) {
      GetFrame()->RemoveGeom(geoms[i]);
    }
  }
  Body::SetFrame(f);
  if(f) {
    for(unsigned int i = 0; i < geoms.size(); i++) {
      f->AddGeom(geoms[i]);
    }
  }
}

void ModelBody::TriMeshUpdateLastPos(void) {
  /* ODE tri mesh likes to know our old position. */
  const dReal* r = dGeomGetRotation(geoms[0]);
  vector3d pos = GetPosition();
  dReal* t = triMeshTrans + 16*triMeshLastMatrixIndex;
  t[ 0] = r[0]; t[1] = r[1]; t[ 2] = r[ 2]; t[ 3] = 0;
  t[ 4] = r[4]; t[5] = r[5]; t[ 6] = r[ 6]; t[ 7] = 0;
  t[ 8] = r[8]; t[9] = r[9]; t[10] = r[10]; t[11] = 0;
  t[12] = pos.x; t[13] = pos.y; t[14] = pos.z; t[15] = 1;
  triMeshLastMatrixIndex = !triMeshLastMatrixIndex;
  for(unsigned int i = 0; i < geoms.size(); i++) {
    dGeomTriMeshSetLastTransform(geoms[i], *(dMatrix4*)(triMeshTrans + 16*triMeshLastMatrixIndex));
  }
}

void ModelBody::RenderSbreModel(const Frame* camFrame, int model, ObjParams* params) {
  glPushMatrix();

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  /* TODO Reduce. */
  glPushAttrib(GL_ALL_ATTRIB_BITS);
  {
    GLfloat lightCol[4];
    glGetLightfv(GL_LIGHT0, GL_DIFFUSE, lightCol);
    lightCol[3] = 0;

    GLfloat lightDir[4];
    glGetLightfv(GL_LIGHT0, GL_POSITION, lightDir);
    lightDir[2] = -lightDir[2];
    sbreSetDirLight(lightCol, lightDir);
  }

  sbreSetViewport(L3D::GetScrWidth(), L3D::GetScrHeight(), L3D::GetScrWidth()
                  *0.5, 5.0f, 100000.0f, 0.0f, 1.0f);
  vector3d pos = GetPositionRelTo(camFrame);
  pos = L3D::world_view->viewingRotation*pos;
  Vector p; p.x = pos.x; p.y = pos.y; p.z = -pos.z;
  matrix4x4d rot;
  rot.LoadFromOdeMatrix(dGeomGetRotation(geoms[0]));
  rot = L3D::world_view->viewingRotation * rot;
  Matrix m;
  m.x1 =  rot[0]; m.x2 =  rot[4]; m.x3 = -rot[8];
  m.y1 =  rot[1]; m.y2 =  rot[5]; m.y3 = -rot[9];
  m.z1 = -rot[2]; m.z2 = -rot[6]; m.z3 = rot[10];
  
  sbreRenderModel(&p, &m, model, params);

  glPopAttrib();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
}

